from pydantic import BaseModel, AwareDatetime, Field, AliasChoices
from exploitfarm.models.dbtypes import AttackGroupID, UnHashedClientID, ExploitID, ExploitSourceID, ClientID
from exploitfarm.utils import json_like
from exploitfarm.models.enums import GroupStatus, GroupEventRequestType, GroupEventResponseType
from typing import Literal
from pydantic import PositiveInt
from uuid import UUID
from pydantic import field_validator

class AddGroupForm(BaseModel):
    name: str
    exploit: str
    created_by: UnHashedClientID
    commit: ExploitSourceID|Literal["latest"] = "latest"
    def db_data(self):
        d = json_like(self, convert_keys={
            "created_by": "created_by_id",
            "exploit": "exploit_id",
            "commit": "commit_id"
        })
        if d["commit_id"] == "latest":
            d["commit_id"] = None
        return d

class EditGroupForm(BaseModel):
    name: str

class GroupDTO(BaseModel):
    id: AttackGroupID
    name: str
    members: list[str] = []
    exploit: ExploitID = Field(validation_alias=AliasChoices('exploit_id'))
    last_attack_at: AwareDatetime|None = None
    status: GroupStatus = GroupStatus.inactive
    commit: UUID|Literal["latest"] = Field("latest", validation_alias=AliasChoices('commit_id'))
    
    class Config:
        validate_assignment = True

    @field_validator('commit', mode='before')
    def commit_none_default(cls, commit):
        return commit or 'latest'

class __JoinRequest(BaseModel):
    group_id: AttackGroupID
    queue_size: PositiveInt

class JoinRequest(__JoinRequest):
    client: UnHashedClientID

class JoinRequestClient(__JoinRequest):
    client: UUID

class JoinRequestResponse(BaseModel):
    timeout: int
    deadline: AwareDatetime
    running: bool

class GroupRequestEvent(BaseModel):
    group_id: AttackGroupID
    event: GroupEventRequestType
    data: dict|None = None

class __GroupResponseEvent(BaseModel):
    group_id: AttackGroupID
    event: GroupEventResponseType
    data: dict|None = None

class GroupResponseEventClient(__GroupResponseEvent):
    client: ClientID

class GroupResponseEvent(__GroupResponseEvent):
    client: UnHashedClientID
